################################################################################
# Google Latitude Heat Map 
#
#     Dale Lane (dale.lane@gmail.com)
#           first written 28 May 2010
#           updated 8 May 2011
################################################################################ 
#  needs:
#    KML file exported from Google Latitude
################################################################################
# parses KML downloaded from http://www.google.com/latitude/
#    and generates JavaScript to make a heatmap
################################################################################
#  warning and disclaimer: 
#    Provided as-is. 
#    This was a very quick bit of fun on a quiet evening - it's almost certainly
#     got a ton of bugs and misunderstandings in it. Take it's results with a 
#     pinch of salt!
################################################################################
# imports
import xml.parsers.expat

################################################################################
#
# VALUES TO TWEAK THE BEHAVIOUR OF THE SCRIPT
#
################################################################################

LATLON_ACCURACY = 8        # how many decimal places should we round the lat/lon
                           #  values to?
                           # reduce this to group similar locations together into
                           #  a single larger group

BOUNDS_LAT_MAX = 90        # range of latitude to include
                           #  set to 90 to get everything
BOUNDS_LAT_MIN = -90       # range of latitude to include
                           #  set to -90 to get everything
BOUNDS_LON_MAX = 180       # range of longitude to include
                           #  set to 180 to get everything
BOUNDS_LON_MIN = -180      # range of longitude to include
                           #  set to -180 to get everything

# for example, in some of my maps, I set the bounds to draw a map of 
#   locations in and around Eastleigh
# BOUNDS_LAT_MAX = 51.033190218589546
# BOUNDS_LAT_MIN = 50.95929172950454
# BOUNDS_LON_MAX = -1.3039398193359375
# BOUNDS_LON_MIN = -1.4261627197265625
BOUNDS_LAT_MAX = 57
BOUNDS_LAT_MIN = 50.1
BOUNDS_LON_MAX = 1.4
BOUNDS_LON_MIN = -5

#BOUNDS_LAT_MAX = 51.44
#BOUNDS_LAT_MIN = 51.404
#BOUNDS_LON_MAX = -1.25
#BOUNDS_LON_MIN = -1.3


ADJUST_DENSITY_FOR_TIME = True   # False - we add every location to the map once
                                 #          for every time it was recorded by 
                                 #          Google Latitude
                                 # True - we add each location to the map a number
                                 #          of times that is proportional to the 
                                 #          total amount of time spent there

DENSITY_ADJUST_HOURS = 24    # if ADJUST_DENSITY_FOR_TIME is True, then we will 
                             #  add each location to the heatmap once for this 
                             #  number of hours spent at the location
                             # e.g. if this number is 12, we will add the 
                             #  location to the heatmap once for every 12 hours
                             #  spent there in total

GOOGLE_LATITUDE_FILE = "kml"   # file name of KML from Google Latitude


################################################################################
# globals
################################################################################
# points we will plot on the heat map
pts = []


class MyKMLParser:
  
    # which XML element are we on?  
    currentelement = None

    # the timestamp (in milliseconds since epoch) of the last placemark
    previoustime = None
    # the timestamp (in milliseconds since epoch) of the current placemark
    latesttime = None
    
    # the latitude of the current placemark
    latestlat = None
    # the longitude of the current placemark
    latestlon = None  

    # two-dimensional array of time spent at locations
    # timespent[lat][lon] = totalTimeInMilliseconds
    timespent = {}
    
    # XML parser function - StartElementHandler
    def start_element(self, name, attrs):    
        latestelement = str(name)
        if latestelement == "Data":
            # there are multiple data elements - it's the 'name' attribute which 
            #  identifies these element
            self.currentelement = attrs["name"]
        elif latestelement != "value":
            # value elements are always within a uniquely named element, so 
            #  we ignore the 'value' tag - it's the parent that tells us what
            #  this
            self.currentelement = latestelement

    # XML parser function - EndElementHandler
    def end_element(self, name):  
        global ADJUST_DENSITY_FOR_TIME, pts
        if name == "Placemark":
            if self.latestlat != None and self.latestlon != None:
                if ADJUST_DENSITY_FOR_TIME == True:
                    #
                    # calculate the time difference between this placemark 
                    #  and the previous one
                    #
                    
                    # default to 10 minutes - for no particularly good reason
                    valueMilliseconds = 10 * 60 * 1000
                    if self.previoustime != None:
                        valueMilliseconds = self.previoustime - self.latesttime
                    self.previoustime = self.latesttime
        
                    #
                    # store the time spent in this location
                    #
                    if self.latestlat not in self.timespent:
                        self.timespent[self.latestlat] = {}
                    if self.latestlon not in self.timespent[self.latestlat]:
                        self.timespent[self.latestlat][self.latestlon] = 0
    
                    self.timespent[self.latestlat][self.latestlon] += valueMilliseconds
                else:
                    #
                    # plot this point - we don't care how about how long was 
                    #  spent here
                    #
                    pts.append((self.latestlon, self.latestlat))
        
        if name != "Data":
            self.currentelement = ""

    # XML parser function - CharacterDataHandler
    def char_data(self, data):
        global LATLON_ACCURACY, BOUNDS_LAT_MAX, BOUNDS_LAT_MIN, BOUNDS_LON_MAX, BOUNDS_LON_MIN

        if self.currentelement == "coordinates":
            #
            # location data - looks like <coordinates>-1.355025,50.973697,0</coordinates>
            #  so we split it on commas, and take the first and second value
            #

            # initialise values
            self.latestlat = None
            self.latestlon = None

            tmplatestcoordinates = str(data)
            coordbits = tmplatestcoordinates.split(',')
            try:
                if len(coordbits) == 3:
                    lat = float(coordbits[1])
                    lon = float(coordbits[0])

                    #if lon <= BOUNDS_LON_MAX and lon >= BOUNDS_LON_MIN and lat <= BOUNDS_LAT_MAX and lat >= BOUNDS_LAT_MIN:
                    if lon != 0 and lat != 0:
                        self.latestlat = round(lat, LATLON_ACCURACY)
                        self.latestlon = round(lon, LATLON_ACCURACY)
            except:
                noop = 0

        elif self.currentelement == "timestamp":
            #
            # timestamp - looks like  <Data name="timestamp"><value>1274989863180</value></Data>
            #  the value is in milliseconds since the epoch
            #

            try:
                newlatesttime = int(data)
                # sanity check - is it in the range I expect?
                #  these are hard-coded values for the times I know that I've 
                #   been using latitude
                if newlatesttime > 1254355200000 and newlatesttime < 1275043497000:
                    self.latesttime = newlatesttime
            except Exception, exc:
                noop = 0


################################################################################
#
# parse the KML 
#
################################################################################
parser = MyKMLParser()
p = xml.parsers.expat.ParserCreate()
p.StartElementHandler  = parser.start_element
p.EndElementHandler    = parser.end_element
p.CharacterDataHandler = parser.char_data

p.ParseFile(open(GOOGLE_LATITUDE_FILE))

################################################################################
#
# draw the extracted points on a heat map 
#
################################################################################

# if we are adjusting the density based on how long was spent at locations, 
#   we need to do this now
# if not, we added points to the heat-map store as we parsed the KML, so there
#   is nothing else to do 

if ADJUST_DENSITY_FOR_TIME == True:
    # collect the points to draw
    for lat in parser.timespent:
        for lon in parser.timespent[lat]:
            # how long did we spend at this location?
            #  so that the heat map reflects the time spent at locations, we 
            #   add a point multiple times 
            timeSpent = parser.timespent[lat][lon] / (1000 * 60 * 60 * DENSITY_ADJUST_HOURS)
            if timeSpent > 0:
                for n in range(int(timeSpent)):
                    pts.append((lon, lat))
            else:
                pts.append((lon, lat))

# draw the heatmap 
print "var coordinates = ["
for point in pts[:-1]:
    print "[ " + str(point[1]) + ", " + str(point[0]) + " ],"
print "[ " + str(pts[-1][1]) + ", " + str(pts[-1][0]) + " ]"
print "];"

